from enum import unique, Enum


@unique
class Ex(Enum):
    # found
    FOUND = "found"

    # text
    TEXT = "text"
    FULL_TEXT = "full_text"

    # tag
    TAG = "tag"
    ETREE = "etree"
    TAG_AS_STRING = "tag_as_string"

    # href
    HREF = "href"
    HREF_QUERY = "href_query"
    HREF_DOMAIN = "href_domain"
    HREF_ENDPOINT = "href_endpoint"
    HREF_BASE_DOMAIN = "href_base_domain"
    HREF_QUERY_PARAMS = "href_query_params"
    HREF_ENDPOINT_WITH_QUERY = "href_endpoint_with_query"


__all__ = [
    "Ex"
]



from . import utils
from lxml import etree, html as lxml_html
from .classes import Ex, BlueMoss, Range, JsonifyWithTag, PrettyDict


def extract(moss: BlueMoss, html: str) -> any:
    return _extract(moss, lxml_html.fromstring(html), 0)


def _extract(moss: BlueMoss, root: etree.Element, level: int) -> any:
    if moss.no_path:
        elems = [root]
    elif not (elems := root.xpath(moss.full_path)):
        if moss.extract == Ex.FOUND:
            return False
        return moss.transform(None if isinstance(moss.filter, int) else [])
    try:
        if isinstance(moss.filter, int):
            elems = [elems[moss.filter]]
        elif isinstance(moss.filter, Range):
            elems = moss.filter.filter(elems)
    except IndexError:
        return moss.transform(None if isinstance(moss.filter, int) else [])
    if moss.nodes:
        if moss.find_single:
            val = moss.transform(_build_target_instance(moss, elems[0], level))
        else:
            val = moss.transform([
                _build_target_instance(moss, elem, level)
                for elem in elems
            ])
    elif moss.find_single:
        val = moss.transform(_extract_from_leaf_node(moss, elems[0]))
    else:
        val = moss.transform([
            _extract_from_leaf_node(moss, elem)
            for elem in elems
        ])
    if moss.key is not None and level == 0:
        return {moss.key: val}
    return val


def _build_target_instance(moss: BlueMoss, elem: etree.Element, level: int):
    if moss.target is None:
        if moss.keys_in_nodes:
            return PrettyDict({node.key: _extract(node, elem, level+1) for node in moss.nodes})
        return [_extract(node, elem, level+1) for node in moss.nodes]
    values: dict[str, any] = {}
    params_with_defaults: set[str] = utils.get_class_params_with_default_value(moss.target)
    for node in moss.nodes:
        val: any = _extract(node, elem, level+1)
        if not (val is None and node.key in params_with_defaults):
            values[node.key] = val
    if issubclass(moss.target, JsonifyWithTag):
        values |= {"_tag": elem}
    # noinspection PyArgumentList
    return moss.target(**values)


def _extract_from_leaf_node(moss: BlueMoss, elem: etree.Element) -> any:
    if type(elem).__name__ == '_ElementUnicodeResult':
        return str(elem)
    if isinstance(moss.extract, str):
        return elem.get(moss.extract)
    match moss.extract:
        case None | Ex.ETREE:
            return elem
        case Ex.FOUND:
            return elem is not None
        case Ex.TEXT:
            return elem.text.strip()
        case Ex.FULL_TEXT:
            return utils.clean_text(elem.text_content().strip())
        case Ex.TAG:
            return utils.etree_to_bs4(elem)
        case Ex.TAG_AS_STRING:
            return utils.etree_to_bs4(elem).prettify()
        case Ex.HREF:
            return elem.get("href")
        case Ex.HREF_QUERY:
            return utils.get_url_query(elem.get("href"))
        case Ex.HREF_DOMAIN:
            return utils.get_domain(elem.get("href"))
        case Ex.HREF_ENDPOINT:
            return utils.get_endpoint(elem.get("href"))
        case Ex.HREF_BASE_DOMAIN:
            return utils.get_base_domain(elem.get("href"))
        case Ex.HREF_QUERY_PARAMS:
            return utils.get_url_query_params(elem.get("href"))
        case Ex.HREF_ENDPOINT_WITH_QUERY:
            return utils.get_endpoint_with_query(elem.get("href"))



def get_infix(text: str, prefix: str, suffix: str) -> str | None:
    """ @return an infix within @param text given a specific prefix and a specific suffix. """
    if not text:
        return
    idx: int = text.find(prefix)
    if idx == -1:
        return
    res: str = text[idx+len(prefix):]
    if not suffix:
        return res
    idx = res.find(suffix)
    if idx == -1:
        return
    return res[:idx]


def clean_text(text: str | None) -> str | None:
    if text is None:
        return
    text = text.strip()
    while "  " in text:
        text = text.replace("  ", " ")
    while "\n " in text:
        text = text.strip().replace("\n ", "\n")
    while " \n" in text:
        text = text.strip().replace(" \n", "\n")
    while "\n\n" in text:
        text = text.strip().replace("\n\n", "\n")
    return text


__all__ = [
    "get_infix",
    "clean_text"
]



import inspect
from functools import cache
from dataclasses import fields, Field, MISSING, is_dataclass


@cache
def get_class_init_params(cls) -> set[str]:
    assert inspect.isclass(cls)
    return set(inspect.getfullargspec(cls.__init__).args[1:])


@cache
def get_class_params_with_default_value(cls) -> set[str]:
    if is_dataclass(cls):
        return {
            f.name for f in fields(cls)
            if f.default != MISSING or f.default_factory != MISSING
        }
    return {
        name for name, param in inspect.signature(cls.__init__).parameters.items()
        if param.default != inspect.Parameter.empty
    }


def _get_default_value(f: Field) -> any:
    if f.default != MISSING:
        return f.default
    if f.default_factory != MISSING:
        return f.default_factory()


__all__ = [
    "get_class_init_params",
    "get_class_params_with_default_value"
]



import abc
from enum import Enum
from json import dumps
from bs4 import BeautifulSoup
from dataclasses import dataclass
from datetime import datetime, date
from collections import OrderedDict


class PrettyDict(dict):
    def __str__(self) -> str:
        return '{\n' + ',\n'.join(f'    {repr(k)}: {repr(v)}' for k, v in self.items()) + '\n}'


@dataclass
class Jsonify(abc.ABC):
    """
    An abstract dataclass to provide a .dict and a .json property, in order to turn any dataclass
    instance into a python dict or json object, while
        1. considering the order of parameters as they are defined within each dataclass that inherits from Dictable.
        2. dropping all protected parameters (those which begin with an underscore).

    Consider a dataclass instance p of type Profile(header: Header, pages: list[list[[Page]])
    while Profile, Header and Page are all dataclasses which inherit from Dictable.
    If you now execute .dict on p, the method will not only dictify @param p.header,
    but also all Page instances within the nested list of lists of @param p.pages.
    """

    def __init__(self):
        self.__dataclass_fields__ = None

    def __post_init__(self):
        pass

    def __str__(self) -> str:
        return self.json

    @property
    def json(self) -> str:
        return dumps(self.dict, indent=4)

    @property
    def dict(self) -> OrderedDict:
        d: dict = {k: v for k, v in self.__dict__.items() if not k.startswith("_")}
        return OrderedDict([
            (key, self.dictify(d[key]))
            for key in self.__dataclass_fields__ if key in d
        ])

    def dictify(self, val: any) -> any:
        if isinstance(val, Enum):
            return val.value
        if isinstance(val, datetime):
            return val.strftime("%Y-%m-%d %H:%M:%S")
        if isinstance(val, date):
            return val.strftime("%Y-%m-%d")
        if isinstance(val, Jsonify):
            return val.dict
        if isinstance(val, list):
            return [self.dictify(v) for v in val]
        if isinstance(val, set):
            return {self.dictify(v) for v in val}
        if isinstance(val, OrderedDict):
            return OrderedDict([
                (self.dictify(k), self.dictify(v))
                for (k, v) in val.items()
            ])
        if isinstance(val, dict):
            return {self.dictify(k): self.dictify(v) for k, v in val.items()}
        if hasattr(val, "__dict__"):
            return val.__dict__
        return val


@dataclass
class JsonifyWithTag(Jsonify):
    """
    Some dataclass instances may need access to their source-html-tag.
    Those dataclasses can inherit from DictableWithTag and thus also get the benefits of the Dictable class.
    """
    _tag: BeautifulSoup

    def __post_init__(self):
        super().__post_init__()

    @property
    def tag(self) -> BeautifulSoup:
        return self._tag

    @property
    def source_line(self) -> int:
        """ The line within the source-html-doc in which @param self._tag was found. """
        return self._tag.sourceline


__all__ = [
    "PrettyDict",
    "Jsonify",
    "JsonifyWithTag"
]



from __future__ import annotations
import pytest
from .constants import HIGH_NESTING_LEVEL_HTML as HTML
from src.bluemoss import Root, Range, Ex, InvalidXpathException, extract


def test_valid_paths():
    Root()
    Root("")
    Root("a")
    Root("//a")


def test_invalid_path_1():
    with pytest.raises(InvalidXpathException):
        Root("body///section")


def test_invalid_path_2():
    with pytest.raises(InvalidXpathException):
        Root("//")


def test_invalid_path_3():
    with pytest.raises(InvalidXpathException):
        Root("///a")


def test_non_existent_path():
    path: str = "h4"
    assert extract(Root(path), HTML) is None
    assert extract(Root(path, filter=3), HTML) is None
    assert extract(Root(path, filter=None), HTML) == []


def test_nested_path():
    path: str = "body//section//li/a/../h3"
    assert extract(Root(path, filter=None), HTML) == ["Service 1", "Service 2"]
    assert extract(Root(path), HTML) == "Service 1"
    assert extract(Root(path, filter=0), HTML) == "Service 1"
    assert extract(Root(path, filter=1), HTML) == "Service 2"
    assert extract(Root(path, filter=Range(0)), HTML) == ["Service 1", "Service 2"]
    assert extract(Root(path, filter=Range(1)), HTML) == ["Service 2"]
    assert extract(Root(path, filter=Range(4)), HTML) == []


def test_nested_path_with_tag_attribute_search():
    path: str = "body//section//li/a[@href='/info']/../h3"
    assert extract(Root(path, filter=None), HTML) == ["Service 2"]
    assert extract(Root(path), HTML) == "Service 2"


def test_nested_path_with_text_search():
    path: str = "html//section/ul//h3[text()='Service 1']/../a"
    assert extract(Root(path), HTML) == "Learn More"
    assert extract(Root(path, extract=Ex.HREF_ENDPOINT), HTML) == "/learn-more"



from __future__ import annotations
from bs4 import BeautifulSoup
from src.bluemoss import utils
from datetime import date, datetime
from dataclasses import dataclass, field
from lxml import html as lxml_html, etree


@dataclass
class PersonDataClass:
    name: str
    birthday: date
    address: str | None = field(default=None)
    friends: list[PersonDataClass] | None = field(default_factory=list)
    last_updated: datetime | None = field(default=None, init=False)


class PersonClass:
    def __init__(
            self,
            name: str,
            birthday: date,
            address: str | None = None,
            friends=None,
    ):
        if friends is None:
            friends = []
        self.name = name
        self.birthday = birthday
        self.address = address
        self.friends = [] if friends is None else friends
        self.last_updated = datetime.now()


CLASS_INIT_PARAMS: set[str] = {"name", "birthday", "address", "friends"}


# general.py
def test_class_get_init_params():
    assert utils.get_class_init_params(PersonDataClass) == CLASS_INIT_PARAMS
    assert utils.get_class_init_params(PersonClass) == CLASS_INIT_PARAMS


def test_get_class_params_with_default_value():
    assert utils.get_class_params_with_default_value(PersonDataClass) == {"address", "friends", "last_updated"}
    assert utils.get_class_params_with_default_value(PersonClass) == {"address", "friends"}


# text.py
def test_clean_text():
    text: str = """
        Hello world!
                        Here.
        
        This is a test.
        
        
        
            Hello world!
            
            
        Test.
            Hello world!
        
              
    """
    assert utils.clean_text(text) == \
        "Hello world!\nHere.\nThis is a test.\nHello world!\nTest.\nHello world!"
    assert utils.clean_text(None) is None


def test_get_infix():
    text: str = "    Hello world!  "
    assert utils.get_infix(text, "He", "orld!") == "llo w"
    assert utils.get_infix("", "He", "orld!") is None
    assert utils.get_infix(text, "NonExistentPrefix", "NonExistentSuffix") is None


# html.py
def test_etree_to_bs4():
    html: str = "<html><body><p>Hello world!</p></body></html>"
    soup: BeautifulSoup = utils.etree_to_bs4(lxml_html.fromstring(html))
    assert isinstance(soup, BeautifulSoup)
    assert str(soup) == html


def test_remove_tags():
    html: str = """<html><body><p>Hello world!</p><h1>Headline</h1><p>This is code.</p></body></html>"""

    for (tags_to_remove, expected_html) in [
        ("p", "<html><body><h1>Headline</h1></body></html>"),
        ("body", "<html></html>"),
        ("html", "")
    ]:
        soup = BeautifulSoup(html, "html.parser")
        utils.remove_tags(soup, tags_to_remove)
        assert str(soup) == expected_html


def test_etree_to_string():
    html: str = "<html><body><p>Hello world!</p></body></html>"
    elem: etree.Element = lxml_html.fromstring(html)
    assert utils.etree_to_string(elem) == html


# url.py

URLS: list[str | None] = [
    "https://example.com/path/to/resource",
    "https://sub1.example.net/resource",
    "https://sub1.sub2.example.org/path/to/resource2?query=val",
    "https://www.sub3.example.com/",
    "https://example2.org/path/to/resource3?p=v&q=w&p=z&a=b",
    "    ",
    "",
    None
]


def test_get_base_domain():
    base_domains: list[str] = [
        "example.com",
        "example.net",
        "example.org",
        "example.com",
        "example2.org",
        None,
        None,
        None
    ]
    for i, url in enumerate(URLS):
        assert utils.get_base_domain(url) == base_domains[i]


def test_get_domain():
    domains: list[str] = [
        "example.com",
        "sub1.example.net",
        "sub1.sub2.example.org",
        "sub3.example.com",
        "example2.org",
        None,
        None,
        None
    ]
    for i, url in enumerate(URLS):
        assert utils.get_domain(url) == domains[i]


def test_get_endpoint():
    endpoints: list[str] = [
        "/path/to/resource",
        "/resource",
        "/path/to/resource2",
        None,
        "/path/to/resource3",
        None,
        None,
        None
    ]
    for i, url in enumerate(URLS):
        assert utils.get_endpoint(url) == endpoints[i]


def test_get_url_query():
    queries: list[str | None] = [
        None,
        None,
        "query=val",
        None,
        "p=v&q=w&p=z&a=b",
        None,
        None,
        None
    ]
    for i, url in enumerate(URLS):
        assert queries[i] == utils.get_url_query(url)


def test_get_endpoint_with_query():
    endpoints_with_query: list[str] = [
        "/path/to/resource",
        "/resource",
        "/path/to/resource2?query=val",
        None,
        "/path/to/resource3?p=v&q=w&p=z&a=b",
        None,
        None,
        None
    ]
    for i, url in enumerate(URLS):
        assert endpoints_with_query[i] == utils.get_endpoint_with_query(url)


def test_get_url_query_params():
    query_params: list[dict[str, str]] = [
        {},
        {},
        {"query": ["val"]},
        {},
        {"p": ["v", "z"], "q": ["w"], "a": ["b"]},
        {},
        {},
        {}
    ]
    for i, url in enumerate(URLS):
        assert query_params[i] == utils.get_url_query_params(url)



from __future__ import annotations
import builtins
from .utils import ClassType
from .extract import Ex
from .range import Range
from inspect import isclass
from functools import cached_property
from dataclasses import dataclass, field
from ..utils import is_valid_xpath, get_class_init_params


@dataclass(frozen=True)
class BlueMoss:
    path: str = ""
    path_prefix: str = ""
    key: str | None = None
    filter: int | None | Range = 0
    target: ClassType | None = None
    extract: Ex | str = Ex.FULL_TEXT
    transform: callable = lambda x: x
    nodes: list[Node] = field(default_factory=list)

    @property
    def no_path(self) -> bool:
        return self.path == ""

    @property
    def find_single(self) -> bool:
        return isinstance(self.filter, int)
    
    @cached_property
    def full_path(self):
        return f"{self.path_prefix}{self.path}"

    @cached_property
    def keys_in_nodes(self) -> set[str]:
        return {c.key for c in self.nodes if c.key is not None}

    def __post_init__(self):
        if not (self.no_path or is_valid_xpath(self.full_path)):
            raise InvalidXpathException(self)

        if not (
                all([c.key is not None for c in self.nodes]) or
                all([c.key is None for c in self.nodes])
        ):
            raise PartialKeysException()
            
        if self.target is None:
            return

        if not isclass(self.target) or self.target.__name__ in dir(builtins):
            raise InvalidTargetTypeException(self)

        if not self.keys_in_nodes.issubset(get_class_init_params(self.target)):
            raise InvalidKeysForTargetException(self)


@dataclass(frozen=True)
class Root(BlueMoss):
    path_prefix: str = field(default="//", init=False)


@dataclass(frozen=True)
class Node(BlueMoss):
    path_prefix: str = field(default=".//", init=False)


class InvalidXpathException(Exception):
    def __init__(self, moss: BlueMoss):
        message: str = (
            f"\n{moss.full_path} seems to be an invalid xpath. "
            f"Feel free to use ChatGPT to check if your path is compatible with the XPath 1.0 syntax. "
            f"Note that xpath queries using XPath syntax of any version higher than 1.0 are not supported."
        )
        super().__init__(message)


class PartialKeysException(Exception):
    def __init__(self):
        message: str = (
            f"\n\nSome Node instances in your nodes list have a key attribute set while others don't."
            f"\nYou can either provide a list of nodes where ALL instances set a key, or NO instances do."
        )
        super().__init__(message)


class InvalidTargetTypeException(Exception):
    def __init__(self, moss: BlueMoss):
        message: str = (
            f"\n\nThe type of your target '{str(moss.target)}' is not a custom class nor a dataclass."
            f"\nMake sure the target either refers to a class or a dataclass."
            f"\n\nPro Tips:"
            f"\n1) If you want your target to be a DICT, "
            f"then you don't have to set the 'target' parameter. All you have to do is to provide keys "
            f"with all Node instances in your 'nodes' list."
            f"\n2) If you want your target to be a LIST, then you also don't need to set your 'target' parameter. "
            f"Just provide the list of Node instances in your 'nodes' list "
            f"without any of the Node instances having set the 'key' parameter."
        )
        super().__init__(message)


class InvalidKeysForTargetException(Exception):
    def __init__(self, moss: BlueMoss):
        target_fields: set[str] = get_class_init_params(moss.target)
        invalid_fields: set[str] = {field for field in moss.keys_in_nodes if field not in target_fields}
        message: str = (
            f"A Node instance in your 'nodes' list defines a key that is no valid init "
            f"parameter for your target {str(moss.target)}: {invalid_fields.pop()}"
        )
        super().__init__(message)


__all__ = [
    "BlueMoss",
    "Root",
    "Node",
    "InvalidXpathException",
    "PartialKeysException",
    "InvalidTargetTypeException",
    "InvalidKeysForTargetException"
]



import pytest
from dataclasses import dataclass
from .constants import FOUR_DIVS_HTML as HTML
from src.bluemoss import Root, Node, Range, extract, EqualIndicesException


def test_find_first():
    expected_output: str = "Hello 1"
    assert extract(Root("div"), HTML) == expected_output
    assert extract(Root("html//div"), HTML) == expected_output


def test_find_last():
    moss = Root("div", filter=-1)
    assert extract(moss, HTML) == "Hello 4"


def test_find_range():
    # first two
    moss = Root("div", filter=Range(0, 2))
    assert extract(moss, HTML) == ["Hello 1", "Hello 2"]

    # middle two
    moss = Root("div", filter=Range(1, -1))
    assert extract(moss, HTML) == ["Hello 2", "Hello 3"]


def test_find_range_reversed():
    # first two reversed
    moss = Root("div", filter=Range(0, 2, reverse=True))
    assert extract(moss, HTML) == ["Hello 2", "Hello 1"]

    # middle two reversed
    moss = Root("div", filter=Range(1, -1, True))
    assert extract(moss, HTML) == ["Hello 3", "Hello 2"]


def test_find_all():
    moss = Root("div", filter=None)
    assert extract(moss, HTML) == [f"Hello {i}" for i in range(1, 5)]


def test_find_all_reversed():
    moss = Root("div", filter=Range(0, None, True))
    assert extract(moss, HTML) == [f"Hello {i}" for i in range(4, 0, -1)]


def test_find_range_with_dict_result():
    # first two
    moss = Root(
        path="html",
        nodes=[
            Node("div", key="1"),
            Node("div[2]", key="2")
        ]
    )
    assert extract(moss, HTML) == {"1": "Hello 1", "2": "Hello 2"}

    # middle two
    li: list[Root] = [
        Root(
            "html",
            nodes=[
                Node("div", key="1", filter=1),
                Node("div", key="2", filter=2)  # Node("div[3]", key="2")
            ]
        ),
        Root(
            "html",
            nodes=[
                Node("div[2]", key="1"),
                Node("div[3]", key="2")
            ]
        )
    ]
    for moss in li:
        assert extract(moss, HTML) == {"1": "Hello 2", "2": "Hello 3"}


def test_find_range_with_class_result():
    @dataclass
    class MyClass:
        a: str
        b: str

        def __eq__(self, other) -> bool:
            return self.a == other.a and self.b == other.b

    # first two
    moss = Root(
        "html",
        target=MyClass,
        nodes=[
            Node("div", key="a"),
            Node("div[2]", key="b")
        ]
    )
    assert extract(moss, HTML) == MyClass(a="Hello 1", b="Hello 2")

    # middle two
    li: list[Root] = [
        Root(
            "html",
            target=MyClass,
            nodes=[
                Node("div", key="a", filter=1),
                Node("div", key="b", filter=2)
            ]
        ),
        Root(
            "html",
            target=MyClass,
            nodes=[
                Node("div[2]", key="a"),
                Node("div[3]", key="b")
            ]
        )
    ]
    for moss in li:
        assert extract(moss, HTML) == MyClass(a="Hello 2", b="Hello 3")


def test_bad_indexing():
    # 1) find many
    moss = Root("div", filter=Range(5))
    assert extract(moss, HTML) == []

    # 2) find single & valid (but out of bound) indexing for Range object
    moss = Root("div", filter=5)
    assert extract(moss, HTML) is None

    # 3) find single & valid indexing for Range object
    moss = Root("div", filter=Range(3, -1))
    assert extract(moss, HTML) == []

    # 4) invalid indexing for Range object
    with pytest.raises(EqualIndicesException):
        Root("div", filter=Range(5, 5))



from urllib.parse import urlparse, parse_qs


def get_base_domain(url: str | None) -> str | None:
    """ @return: the top-level-domain of the given @param url. """
    domain: str | None = get_domain(url)
    return ".".join(domain.split(".")[-2:]) if domain else None


def get_domain(url: str | None) -> str | None:
    """ @return: the full domain of the given @param url. """
    url: str | None = url.strip() if url else None
    if not url:
        return
    domain = urlparse(url).netloc
    return domain.replace("www.", "", 1)


def get_endpoint(url: str | None) -> str | None:
    """ @return: the endpoint of the given @param url ."""
    url: str | None = url.strip() if url else None
    if not url:
        return
    endpoint: str | None = urlparse(url).path if url else None
    if endpoint and endpoint.endswith("/"):
        endpoint = endpoint[:-1]
    if endpoint:
        return endpoint


def get_url_query(url: str | None) -> str | None:
    url: str | None = url.strip() if url else None
    if not url:
        return
    query: str = urlparse(url).query
    return query if query else None


def get_endpoint_with_query(url: str | None) -> str | None:
    if not (endpoint := get_endpoint(url)):
        return
    if not (query := get_url_query(url)):
        return endpoint
    return f"{endpoint}?{query}"


def get_url_query_params(url: str | None) -> dict:
    return parse_qs(urlparse(url).query) if url else {}


__all__ = [
    "get_base_domain",
    "get_domain",
    "get_endpoint",
    "get_url_query",
    "get_endpoint_with_query",
    "get_url_query_params"
]



from lxml import etree
from bs4 import BeautifulSoup
from src.bluemoss import Root, Ex, extract
from .constants import WITH_LINKS_HTML as HTML


def test_found_extraction():
    # tag can be found
    moss = Root("div[@class='id_1']", extract=Ex.FOUND)
    assert extract(moss, HTML) is True

    # tag cannot be found
    moss = Root("div[@class='id_3']", extract=Ex.FOUND)
    assert extract(moss, HTML) is False


def test_text_extraction():
    moss = Root("p", extract=Ex.TEXT)
    assert extract(moss, HTML) == "Lorem 1"

    moss = Root("div", extract=Ex.TEXT)
    assert extract(moss, HTML) == "Ipsum 2"


def test_full_text_extraction():
    moss = Root("div", extract=Ex.FULL_TEXT)
    assert extract(moss, HTML) == "Ipsum 2\nLorem 2\nLink 2"


def test_tag_extraction():
    moss = Root("div", extract=Ex.TAG)
    tag: BeautifulSoup = extract(moss, HTML)
    assert isinstance(tag, BeautifulSoup)

    moss = Root("div")
    assert extract(moss, tag.prettify()) == "Ipsum 2\nLorem 2\nLink 2"


def test_etree_extraction():
    # 1) extract the found tag as an etree._Element instance
    moss = Root("div", extract=Ex.ETREE)
    elem: etree._Element = extract(moss, HTML)
    assert isinstance(elem, etree._Element)

    # 2) transform the etree._Element into a html string and test the full-text-extraction on it
    html: str = etree.tostring(elem, method="html")
    moss = Root("div")
    assert extract(moss, html) == "Ipsum 2\nLorem 2\nLink 2"


def test_tag_as_string_extraction():
    moss = Root("div", extract=Ex.TAG_AS_STRING)
    html: str = extract(moss, HTML)
    assert isinstance(html, str)
    moss = Root("div")
    assert extract(moss, html) == "Ipsum 2\nLorem 2\nLink 2"


def test_href_extraction():
    moss = Root("a", filter=2, extract=Ex.HREF)
    assert extract(moss, HTML) == "https://www.nvidia.com/link3?p=3&q=3"


def test_href_query_extraction():
    moss = Root("a", extract=Ex.HREF_QUERY)
    assert extract(moss, HTML) == "p=1&q=1"


def test_href_domain_extraction():
    moss = Root("a", filter=3, extract=Ex.HREF_DOMAIN)
    assert extract(moss, HTML) == "chat.openai.com"


def test_href_base_domain_extraction():
    moss = Root("a", filter=3, extract=Ex.HREF_BASE_DOMAIN)
    assert extract(moss, HTML) == "openai.com"


def test_href_endpoint_extraction():
    moss = Root("a", filter=3, extract=Ex.HREF_ENDPOINT)
    assert extract(moss, HTML) == "/page4/link4"


def test_href_endpoint_with_query_extraction():
    moss = Root("a", filter=3, extract=Ex.HREF_ENDPOINT_WITH_QUERY)
    assert extract(moss, HTML) == "/page4/link4?p=4&q=4&p=5"


def test_href_query_params_extraction():
    moss = Root("a", filter=3, extract=Ex.HREF_QUERY_PARAMS)
    assert extract(moss, HTML) == {"p": ["4", "5"], "q": ["4"]}



from __future__ import annotations
import pytest
from dataclasses import dataclass
from src.bluemoss.classes.dict import Jsonify
from .constants import WITH_LINKS_HTML as HTML
from src.bluemoss.utils import url as url_utils
from src.bluemoss import BlueMoss, Root, Node, Ex, extract, InvalidTargetTypeException, InvalidKeysForTargetException


@dataclass
class Links(Jsonify):
    links: list[Link]


@dataclass
class Link(Jsonify):
    url: str

    @property
    def domain(self) -> str:
        return url_utils.get_domain(self.url)

    @property
    def base_domain(self) -> str:
        return url_utils.get_base_domain(self.url)

    @property
    def endpoint(self) -> str:
        return url_utils.get_endpoint(self.url)

    @property
    def dict(self) -> dict:
        return {
            "url": self.url,
            "domain": self.domain,
            "endpoint": self.endpoint,
            "base_domain": self.base_domain
        }


MOSS_WITH_TARGET: BlueMoss = Root(
    target=Links,
    nodes=[
        Node(
            path="a",
            key="links",
            target=Link,
            filter=None,
            nodes=[
                Node(
                    key="url",
                    extract=Ex.HREF
                )
            ]
        )
    ]
)


MOSS_WITHOUT_TARGET: BlueMoss = Root(
    path="a",
    key="links",
    filter=None,
    nodes=[
        Node(
            key="url",
            extract=Ex.HREF
        ),
        Node(
            key="domain",
            extract=Ex.HREF_DOMAIN
        ),
        Node(
            key="base_domain",
            extract=Ex.HREF_BASE_DOMAIN
        ),
        Node(
            key="endpoint",
            extract=Ex.HREF_ENDPOINT
        )
    ]
)


def test_extract_equality():
    assert extract(MOSS_WITH_TARGET, HTML).dict == extract(MOSS_WITHOUT_TARGET, HTML)


def test_invalid_target_type():
    with pytest.raises(InvalidTargetTypeException):
        Root(target=list)
    with pytest.raises(InvalidTargetTypeException):
        Root(target=dict)
    with pytest.raises(InvalidTargetTypeException):
        Root(target=set)


def test_invalid_keys_for_target():
    with pytest.raises(InvalidKeysForTargetException):
        Root(target=Link, nodes=[Node(key="domain")])



from __future__ import annotations
from enum import Enum
from dataclasses import dataclass
from collections import OrderedDict
from datetime import date, datetime
from src.bluemoss.classes.dict import Jsonify


class Job(Enum):
    ENGINEER = "Engineer"
    FARMER = "Farmer"
    LAWYER = "Lawyer"


@dataclass
class Address(Jsonify):
    city: str
    street: str
    zip_code: str
    country_code: str


@dataclass
class Skill(Jsonify):
    name: str
    started_at: date
    experience_level: int


@dataclass
class Person(Jsonify):
    name: str
    email: str
    birthday: date
    address: Address
    last_contacted: datetime
    job: Job
    skills: list[Skill]
    languages: set[str]
    _last_updated: datetime

    @property
    def age(self) -> int:
        today: date = date.today()
        age = today.year - self.birthday.year
        if (today.month, today.day) < (self.birthday.month, self.birthday.day):
            age -= 1
        return age

    @property
    def dict(self) -> dict:
        return super().dict | {"age": self.age}


address = Address(
    city="Los Angeles",
    street="789 Creative Blvd",
    zip_code="90001",
    country_code="US"
)


skill1 = Skill(
    name="Software Development",
    started_at=date(2010, 6, 1),
    experience_level=5
)


skill2 = Skill(
    name="Database Management",
    started_at=date(2012, 8, 10),
    experience_level=4
)


person = Person(
    name="Charlie Davis",
    email="charlie.davis@example.com",
    birthday=date(1988, 10, 20),
    address=address,
    last_contacted=datetime(2023, 10, 21, 19, 24, 35),
    job=Job.ENGINEER,
    skills=[skill1, skill2],
    languages={"English", "French"},
    _last_updated=datetime.now()
)


expected_dict: OrderedDict = OrderedDict([
    ('name', 'Charlie Davis'),
    ('email', 'charlie.davis@example.com'),
    ('birthday', '1988-10-20'),
    ('address',
        OrderedDict([
            ('city', 'Los Angeles'),
            ('street', '789 Creative Blvd'),
            ('zip_code', '90001'),
            ('country_code', 'US')
        ])
     ),
     ('last_contacted', '2023-10-21 19:24:35'),
     ('job', 'Engineer'),
     ('skills', [
         OrderedDict([
             ('name', 'Software Development'),
             ('started_at', '2010-06-01'),
             ('experience_level', 5)
         ]),
         OrderedDict([
             ('name', 'Database Management'),
             ('started_at', '2012-08-10'),
             ('experience_level', 4)
        ])
     ]),
     ('languages', {'French', 'English'}),
     ('age', 35)
])


def test_person_dict():
    assert person.dict == expected_dict



from dataclasses import dataclass, field


@dataclass(frozen=True)
class Range:
    start_idx: int
    end_idx: int | None = None
    reverse: bool = field(default=False)

    def __post_init__(self):
        if isinstance(self.end_idx, int):
            if self.end_idx == self.start_idx:
                raise EqualIndicesException(self.start_idx)

    def filter(self, li: list) -> list:
        if self.end_idx is None:
            res = li[self.start_idx:]
        else:
            res = li[self.start_idx:self.end_idx]
        return res[::-1] if self.reverse else res


class EqualIndicesException(Exception):
    def __init__(self, index: int):
        message: str = (
            f"\n\nYou have defined a Range instance with start_idx={index} and end_idx={index}."
            f"\nMake sure to provide unequal values for your 'start_idx' and 'end_idx' parameters."
            f"\n\nPro Tips:"
            f"\n1) You can define a Range instance with just one value, "
            f"e.g. Range(2) which will filter all matched html-tags from index 2 onwards."
            f"\n2) An instance like Range(0, 2) will filter for the first two matched html-tags."
        )
        super().__init__(message)


__all__ = [
    "Range",
    "EqualIndicesException"
]



import pytest
from src.bluemoss import Root, Node, PartialKeysException


def test_partial_keys_exception():
    Root(nodes=[])
    Root(nodes=[Node(key="some_key")])
    with pytest.raises(PartialKeysException):
        Root(
            nodes=[
                Node(),
                Node(key="some_key")
            ]
        )



from __future__ import annotations
import pytest
from dataclasses import dataclass
from .constants import HIGH_NESTING_LEVEL_HTML as HTML
from src.bluemoss import Root, Node, Range, Ex, extract


def test_domain_extract_via_transform():
    path: str = "footer//li[contains(text(), 'mail')]/a"
    moss = Root(
        path,
        extract=Ex.HREF,
        transform=lambda mail: mail.split('@')[-1] if mail else None
    )
    assert extract(moss, HTML) == 'example.com'


def test_zip_code_extract_via_transform():
    moss = Root(
        "address//div[contains(@class, 'contact-cell')]/p[2]",
        transform=lambda address: int(address.split(' ')[-1])
    )
    assert extract(moss, HTML) == 12345


def test_attempt_transform_on_none():
    moss = Root("h4", transform=lambda none_val: none_val.split())
    with pytest.raises(AttributeError):
        extract(moss, HTML)


def test_usage_and_non_usage_of_transform_param():
    @dataclass
    class Address:
        line_1: str
        line_2: str

    expected: Address = Address(line_1='123 Main Street', line_2='City, State 12345')
    for moss in [
        Root(
            "address//p",
            filter=Range(1),
            transform=lambda lines: Address(lines[0], lines[1])
        ),
        Root(
            "address",
            target=Address,
            nodes=[
                Node("p", filter=1, key="line_1"),
                Node("p", filter=2, key="line_2")
            ]
        )
    ]:
        assert extract(moss, HTML) == expected



# -*- coding: utf-8 -*-

__title__ = 'bluemoss'
__author__ = 'LucaTabone'
__license__ = 'APACHE'
__copyright__ = 'Copyright 2023-2023 LucaTabone'
__version__ = '0.1.1'

from .classes import (
    BlueMoss, Node, Root, Range, Ex, InvalidTargetTypeException, InvalidKeysForTargetException,
    InvalidXpathException, PartialKeysException, EqualIndicesException, Jsonify, JsonifyWithTag
)
from .extract import extract



from lxml import etree
from bs4 import BeautifulSoup


def is_valid_xpath(xpath_query: str):
    root = etree.Element("root")
    doc = etree.ElementTree(root)
    try:
        doc.xpath(xpath_query)
        return True
    except etree.XPathEvalError:
        return False


def etree_to_bs4(node: etree.Element) -> BeautifulSoup:
    return BeautifulSoup(etree_to_string(node), "html.parser")


def etree_to_string(node: etree.Element) -> str:
    return etree.tostring(node, method="html").decode("utf-8")


def remove_tags(soup: BeautifulSoup, tag_names: list[str]) -> str:
    """
    Removes specific HTML tags from a BeautifulSoup object.

    :param soup: BeautifulSoup object representing a block of HTML
    :param tag_names: Tag names to be removed HTML
    """
    for data in soup(tag_names):
        data.decompose()
    return soup.prettify()


__all__ = [
    "is_valid_xpath",
    "etree_to_bs4",
    "etree_to_string",
    "remove_tags"
]
